import os
import socket
import sys
import http.server
import socketserver
import ipaddress
import struct
import subprocess, platform


def get_ip():
  s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  try:
    # doesn't even have to be reachable
    s.connect(('8.8.8.8', 1))
    IP = s.getsockname()[0]
    print (s.getsockname())
  except:
    IP = '127.0.0.1'
  finally:
    s.close()
  return IP

    
def ping(host):
  # Ping parameters as function of OS
  DETACHED_PROCESS = 0x00000008
  ping_str = "-n 1" if  platform.system().lower()=="windows" else "-c 1"
  args = "ping " + " " + ping_str + " " + host
  need_sh = False if  platform.system().lower()=="windows" else True
  # Ping
  return subprocess.call(args, shell=need_sh, creationflags=DETACHED_PROCESS) == 0

def detect_network() :
  """
  Test if current network behind FW to then open internet
  replaced by is_on_open_network()
  return 
    
  """
  
  if ping('8.8.8.8'):
    return "OPEN"
  else :
    return "INTERNAL"

def is_on_open_network() :
  """
  Test if current network behind FW to then open internet
    
  """
  
  if test_port('8.8.8.8', 53) == True:
    return True
  else :
    return False



def test_port (host, port):
  sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  result = sock.connect_ex((host,port))
  if result == 0: 
    return True # open
  return False

def http_server(port, web_path) :
  os.chdir(web_path)
  Handler = http.server.SimpleHTTPRequestHandler
  httpd = socketserver.TCPServer(("", port), Handler)
  print("HTTP Server started at port : ", port)
  try:
    httpd.serve_forever()
  except KeyboardInterrupt:
    httpd.shutdown()
    sys.exit(0)
  sys.exit(0)  

def wake_on_lan(remote_host_mac):
  remote_host_mac = remote_host_mac.replace(":","")
  BROADCAST_IP = '192.168.1.255'
  DEFAULT_PORT = 9
  
  # Pad the synchronization stream
  data = b'FFFFFFFFFFFF' + (remote_host_mac * 16).encode()
  send_data = b''
  
  # Split up the hex values in pack
  for i in range(0, len(data), 2):
    send_data += struct.pack(b'B', int(data[i: i + 2], 16))

  packet = send_data
  print (packet)
  sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
  sock.connect((BROADCAST_IP, DEFAULT_PORT))
  sock.send(packet)
  sock.close()

  
if __name__ == "__main__":
  wake_on_lan("48:5B:39:3E:02:07")
  """
  print(detect_network())
  while True:
    print (is_on_open_network())
  pass  
  """



import sys, os, time, atexit, signal

class Daemon:
  """
  A generic daemon class.
  Usage: subclass the daemon class and override the run() method.
  """
  def __init__(self, pidfile):  
    self.pidfile = pidfile

  def daemonize(self):
    """Deamonize class. UNIX double fork mechanism."""

    try:
      pid = os.fork()
      if pid > 0:
        # exit first parent
        sys.exit(0)
    except OSError as err:
      sys.stderr.write('fork #1 failed: {0}\n'.format(err))
      sys.exit(1)

    # decouple from parent environment
    os.chdir('/')
    os.setsid()
    os.umask(0)

    # do second fork
    try:
      pid = os.fork()
      if pid > 0:

        # exit from second parent
        sys.exit(0)
    except OSError as err:
      sys.stderr.write('fork #2 failed: {0}\n'.format(err))
      sys.exit(1)

    # redirect standard file descriptors
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(os.devnull, 'r')
    so = open(os.devnull, 'a+')
    se = open(os.devnull, 'a+')

    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    # write pidfile
    atexit.register(self.delpid)

    pid = str(os.getpid())
    with open(self.pidfile,'w+') as f:
      f.write(pid + '\n')

  def delpid(self):
    os.remove(self.pidfile)

  def start(self):
    """Start the daemon."""

    # Check for a pidfile to see if the daemon already runs
    try:
      with open(self.pidfile,'r') as pf:

        pid = int(pf.read().strip())
    except IOError:
      pid = None

    if pid:
      message = "pidfile {0} already exist. Daemon already running?\n"
      sys.stderr.write(message.format(self.pidfile))
      sys.exit(1)

    # Start the daemon
    self.daemonize()
    self.run()

  def stop(self):
    """Stop the daemon."""

    # Get the pid from the pidfile
    try:
      with open(self.pidfile,'r') as pf:
        pid = int(pf.read().strip())
    except IOError:
      pid = None

    if not pid:
      message = "pidfile {0} does not exist. Daemon not running?\n"
      sys.stderr.write(message.format(self.pidfile))
      return # not an error in a restart

    # Try killing the daemon process
    try:
      while True:
        os.kill(pid, signal.SIGTERM)
        time.sleep(0.1)
    except OSError as err:
      e = str(err.args)
      if e.find("No such process") > 0:
        if os.path.exists(self.pidfile):
          os.remove(self.pidfile)
      else:
        print (str(err.args))
        sys.exit(1)

  def restart(self):
    """Restart the daemon."""
    self.stop()
    self.start()


  def status(self):
    # Check for a pidfile to see if the daemon already runs
    try:
      with open(self.pidfile,'r') as pf:
        pid = int(pf.read().strip())
    except IOError:
      pid = None

    if pid:
      message = self.pidfile + " already exist. Daemon already running?\n"
    else:
      message = self.pidfile + " not exist. Daemon is not running?\n"
    
    print (message)


  def run(self):
    """You should override this method when you subclass Daemon.

    It will be called after the process has been daemonized by
    start() or restart()."""
    pass



import os
import numpy as np
import exifread


def get_exif(path):
  with open(path, 'rb') as f:
    tags = exifread.process_file(f)

  return tags


def img2array(img_file) :
  img=np.asarray(Image.open(img_file))
  print (img)
  print(img.shape)
  
  
if __name__== "__main__" :
  print(get_exif("R:/abc.jpg"))



import json
#import .fs_tools as fs_tools

def getdicv(dic, key):
  for k,v in dic.items():
    if k == key:
      return v
  return None
  
  
if __name__ == "__main__": 
  a = {"a.b" : "123"}
  print (getdicv(a,"a.b"))
  pass


import os
import sys
import re
import hashlib
import pathlib
import glob
import gzip
import codecs
import base64
import time


from shutil import copyfile,rmtree
from stat import * 
from py3toolbox.text_tools import re_findall


####################################################################################################

def rm_folder(folder_name, empty=False):
  """
  Remove/Delete Folder
  Usage :
    rm_folder(folder_name, empty=False):
      - empty = True  : empty the folder only, folder still remains
      - empty = False : Remove the entire folder with all sub folder and files.

  Sample :
    rm_folder("R:/temp1", empty=False)

  """
  if empty == False:
    rmtree(folder_name,ignore_errors=True)
  else:
    rmtree(folder_name,ignore_errors=True)
    mk_folder(folder_name=folder_name)

delete_folder = rm_folder
rm_dir        = rm_folder


####################################################################################################

def mk_folder(folder_name) :
  """
    Create New Folder(s)
    Usage : 
      mk_folder(folder_name) 
      - folder_name can be multiple layers, and all of them will be created.

    Sample:  
      mk_folder("r:/temp1/t1/t2/t3") 

  """
  if file_exists(folder_name) == False :
    os.makedirs(folder_name, exist_ok=True)

create_folder = mk_folder
make_folder   = mk_folder
make_dir      = mk_folder
####################################################################################################


def clean_folder(folder_name):
  """
    Clean/Empty Folder

    Usage : 
      clean_folder(folder_name) 
      - folder_name : full folder path

    Sample:  
      clean_folder("r:/temp1/t1/t2") 

  """
  mk_folder(folder_name)
  rm_folder(folder_name, empty=True)


####################################################################################################

def get_file_folder(file_path) :
  """
    Get folder path of given file full path
    Usage : 
      get_file_folder(file_path) 
      - file_path : is the full file path

    Sample:  
      get_file_folder("r:/temp1/t1/t2/t3/file1.txt") => r:/temp1/t1/t2/t3

  """  
  return os.path.dirname(file_path)


####################################################################################################


def get_file_name(file_path) :
  """
    Get file base name from given full path
    Usage : 
      get_file_name(file_path) 
      - file_path : is the full file path

    Sample:  
      get_file_name("r:/temp1/t1/t2/t3/file1.txt") => file1.txt

  """    
  return os.path.basename(file_path)


####################################################################################################

def parse_full_path (file_path):
  """
    Get file base name from given full path
    Usage : 
      get_file_name(file_path) 
      - file_path : is the full file path

    Sample:  
      parse_full_path("r:/temp1/t1/t2/t3/file1.txt") => ("r:/temp1/t1/t2/t3", "file1.txt")

  """      
  path = get_file_folder(file_path)
  name = get_file_name(file_path)
  return (path, name)


####################################################################################################





def file_exists(file_name):
  return os.path.exists(file_name)
exists = file_exists


####################################################################################################
  
def copy_file(src_file,dest_file):
  if file_exists(src_file) :  
    mk_folder(get_file_folder(dest_file))
    copyfile(src_file,dest_file)    





####################################################################################################  
def rm_file(file_name = None, folder = None, regex_pattern=None) :
  """
    Remove file(s), if file_name is list, then remove all of them.
    Remove file based on the regex pattern 
    
    Sample:
      rm_file(file_name = 'r:/temp/1.txt')                      ==> remove R:/temp/1.txt
      rm_file(file_name = ['r:/temp/1.txt', 'r:/temp/2.txt'] )  ==> remove R:/temp/1.txt and R:/temp/2.txt
      
      rm_file(folder = "R:/temp", regex_pattern='\d+\.txt' )    ==> remove R:/temp/{1,2,3...}.txt
      
  """
  
  # if only folder and regex_pattern provided
  if file_name is None and folder is not None and regex_pattern is not None:
    file_name = []
    for f in get_files(folder = folder, type='file'):
      if len(re_findall(regex_pattern, f)) > 0:
        file_name.append(f)
    
  # perform file delete 
  if type(file_name) is list:
    for f in file_name:
      if os.path.isfile(f): 
        os.remove(f)
  else:
    if os.path.isfile(file_name): 
      os.remove(file_name)

del_file = rm_file

####################################################################################################          
        
def write_file(file_name, text, mode='w', is_new = False):
  file_folder = get_file_folder(file_name)
  if file_exists(file_folder) == False: mk_folder(file_folder)
    
  if is_new == True: rm_file(file_name)
  with codecs.open(file_name, mode, 'utf8') as fh:
    fh.write(text)

####################################################################################################  


def write_log(file_name, text, is_new = False):
  write_file(file_name=file_name, text=text + "\n", mode='a', is_new=is_new)


####################################################################################################  


    
def read_file(file_name, return_type='str'):
  if return_type == 'str' :
    #return codecs.open(file_name, 'r', 'utf8').read() -- codecs.open obsolete
    return open(file_name, 'r', encoding='utf-8').read()
    
    
  if return_type == 'list' :
    #return codecs.open(file_name, 'r', 'utf8' ).readlines() -- codecs.open obsolete
    return open(file_name, 'r', encoding='utf-8').readlines()


def read_file_lines(file_object, chunk_size=1024):
    #Lazy function (generator) to read a huge file line by line.
    while True:
      line = file_object.readline()
      if not line:
        break
      yield line



####################################################################################################

    
def get_md5(file_name) :  
  hasher = hashlib.md5()
  with open(file_name, 'rb') as fh:
    buf = fh.read()
    hasher.update(buf)
  return (hasher.hexdigest())

####################################################################################################



def get_md5_txt(text) :  
  return hashlib.md5(text.encode('utf-8')).hexdigest()


####################################################################################################

  
def is_file (file) :
  return os.path.isfile(file) 

def is_folder (file) :
  return os.path.isdir(file) 
  
def check_type(full_path):
  if file_exists(full_path) : 
    if is_file(full_path) : 
      return 'file'
    elif is_folder(full_path) :  
      return 'dir'
  else:
    return None
  
####################################################################################################  
  
  
def scan_fs_tree(folder, ext='', recursive=True, type='A'):
  
  for obj in os.scandir(folder):
    if obj.is_dir(follow_symlinks=False):
      if type.upper()=='A' or type.upper()=='ALL' or type.upper()=='D' or type.upper()=='DIR'  or type.upper()=='FOLDER':
        yield obj
      if recursive  == True: 
        yield from scan_fs_tree(folder=obj.path, ext=ext, recursive=recursive, type=type) 
    else:
      if type.upper()=='A' or type.upper()=='ALL' or type.upper()=='F' or type.upper()=='FILE':
        if ext == '':
          yield obj
        else:
          if obj.name.endswith(ext) :  yield obj
      pass

####################################################################################################


def gen_files(folder, ext='',  recursive=True, type='A'):
  for f in scan_fs_tree(folder=folder,  ext=ext, recursive=recursive, type=type):
    file_path = f.path
    if os.name == 'nt':
      file_path = file_path.replace('\\', '/')
    yield file_path
    
####################################################################################################
def get_files(folder, ext='', recursive=True, type='A') : 
  files = []
  for f in gen_files(folder=folder, ext = ext, recursive=recursive, type=type):
    files.append(f)
  return files

####################################################################################################

def get_file_info(file_name) :
  if file_exists(file_name):
    st = os.stat(file_name)
    return { "file_name": file_name, "size" : int(st[ST_SIZE]), "access_time" : st[ST_ATIME] }
  else:
    return None

####################################################################################################
# 
#   
def get_folder_info(folder, recursive=True ):
  """
    
  
  
  """
  
  total_size  = 0
  total_count = 0
  files = get_files(folder=folder, recursive=recursive)
  total_count = len(files)
  for f in files : 
    total_size += get_file_info(f)["size"]
  
  return (total_count, total_size)  

####################################################################################################

def do_zip(input_file=None,output_file=None,input_data=None, output="file"):
  """
  usage  : 
    zip compress input file or bytes string or string
    output: file  : save as file
            byte  : return bytes string
  
  examples:
    do_zip(input_file="c:/1.txt",output_file="c:/1.zip")
    do_zip(input_file="c:/1.txt",output="byte")

    do_zip(dest_file="c:/1.zip", input_data="test_string",output="file")
    do_zip(input_data="test_string",output="byte")

  """

  assert output == "file" or  output == "byte" , "Wrong output type : should be file or byte"

  # check input 
  if input_file is not None:
    with open(input_file, 'rb') as f_in: 
      input_data = f_in.read()
  elif input_data is not None:
    if type(input_data).__name__ == 'str' :
      input_data =  str.encode(input_data)


  compressed_bytes = gzip.compress(input_data, compresslevel=9)


  if output == "file":
    with open(output_file, 'wb') as f_out: 
      f_out.write(compressed_bytes)
    return output_file
  elif output == "byte":
    return  compressed_bytes

  pass


####################################################################################################

def do_unzip(input_file=None,output_file=None,input_data=None, output="file"):
  """
  usage  : 
    do_unzip decompress (unzip) input file or bytes string or string
    output: file  : save as file
            byte  : return bytes string
  
  examples:
    do_unzip(input_file="c:/1.zip",dest_file="c:/1.bin")
    do_unzip(input_file="c:/1.zip",output="byte")

    do_unzip(output_file="c:/1.zip", input_data="test_string",output="file")
    do_unzip(input_file="test_string",output="byte")

  """

  assert output == "file" or  output == "byte" , "Wring output type : should be file or byte"

  # check input 
  if input_file is not None:
    with open(input_file, 'rb') as f_in: 
      input_data = f_in.read()
  elif input_data is not None:
    if type(input_data).__name__ == 'str' :
      input_data =  str.encode(input_data)

  decompressed_bytes = gzip.decompress(input_data)

  if output == "file":
    with open(output_file, 'wb') as f_out: 
      f_out.write(decompressed_bytes)
    return output_file
  elif output == "byte":
    return  decompressed_bytes

  pass



"""
def unzip(gzfile) :
  fgz = gzip.GzipFile(gzfile, 'rb')
  binary_content = fgz.read()
  fgz.close()
  return binary_content
"""



####################################################################################################

def split_file(input_file, output_folder, volume_size=10):
  """
    Split file to multiple volumes by size 
    
    Sample:
      split_file(input_file='R:/Temp2/disk.bin', output_folder ='R:/Temp2', volume_size=1234)
      join_file(input_file='R:/Temp2/disk.bin',  output_folder ='R:/Temp1')
  
  """
  assert input_file is not None and output_folder is not None, "Error: input_file, output_folder are required."
  
  part_number = 0
  volume_size = volume_size * 1024 * 1024
  
  # cleanup volume files before creating new ones
  rm_file(folder = output_folder, regex_pattern = get_file_name(input_file) + '\.part_\d+')
  
  with open(input_file,"rb") as f:
    chunk = f.read(volume_size)
    while chunk:
      chunk_file = output_folder + '/' + get_file_name(input_file) + '.part_' + str(part_number).zfill(5)
      with open(chunk_file , "wb" ) as chunk_fh:
        chunk_fh.write(chunk)
        chunk_fh.flush()
      part_number += 1
      chunk = f.read(volume_size)
    
  return

####################################################################################################


def join_file(input_file, output_folder , volume_size=None):
  """
    Join files from multiple volumes 
    
    Sample:
      split_file(input_file='R:/Temp2/disk.bin', output_folder ='R:/Temp2', volume_size=1234)
      join_file(input_file='R:/Temp2/disk.bin',  output_folder ='R:/Temp1')
  
  """
  
  assert input_file is not None  or  output_file is not  None, "input_file, output_file are required"
  
  input_folder, file_name = parse_full_path(input_file)
  
  file_list = []
  for f in get_files(folder=input_folder, type='file') :
    if file_name + '.part_' in f:
      file_list.append(f)
  file_list.sort()
  
  output_file = os.path.join(output_folder , file_name)
  
  # cleanup volume files before creating new ones
  rm_file(file_name = os.path.join(output_folder,output_file))
  
  for f in file_list:
    with open(f,"rb") as fh:
      chunk = fh.read()
      with open(output_file , "ab" ) as  output_fh:
        output_fh.write(chunk)
        output_fh.flush()

  print (file_list)
  return
  
####################################################################################################
def compare_files(source, target):
  """
    Compare two files by MD5
    
    Sample:
      compare_files('R:/Temp1/1.txt', 'R:/Temp1/2.txt') 
      - Returns True  : two files identical
      - Returns False : two files different
      
  """
  
  return (get_md5(source) == get_md5(target))


####################################################################################################

def base64_encode(input_file=None,output_file=None,input_data=None, output="file"):
  """
  usage  : 
    base64 encode input file or bytes string or string
    output: file  : save as file
            text  : return bytes string
  
  examples:
    base64_encode(input_file="c:/1.txt",output_file="c:/1.zip")
    base64_encode(input_file="c:/1.txt",output="text")

    base64_encode(dest_file="c:/1.zip", input_data="test_string",output="file")
    base64_encode(input_data="test_string",output="text")

  """
  assert output == "file" or  output == "text" , "Wrong output type : should be file or text"
  
  # check input 
  if input_file is not None:
    with open(input_file, 'rb') as f_in: 
      input_data = f_in.read()
  elif input_data is not None:
    if type(input_data).__name__ == 'str' :
      input_data =  str.encode(input_data)

  encoded_bytes = base64.b64encode(input_data)
 
  if output == "file":
    with open(output_file, 'wb') as f_out: 
      f_out.write(encoded_bytes)
    return output_file
  elif output == "text":
    return  encoded_bytes.decode() 


#################################################################################################### 

def base64_decode(input_file=None,output_file=None,input_data=None, output="file"):
  """
  usage  : 
    base64 decode input file or bytes string or string
    output: file  : save as file
            byte  : return bytes string
  
  examples:
    base64_decode(input_file="c:/1.txt",output_file="c:/1.zip")
    base64_decode(input_file="c:/1.txt",output="byte")

    base64_decode(dest_file="c:/1.zip", input_data="test_string",output="file")
    base64_decode(input_data="test_string",output="byte")

  """
  
  assert output == "file" or  output == "byte" , "Wring output type : should be file or byte"
  input_text = None

  # check input 
  if input_file is not None:
    with open(input_file, 'rb') as f_in: 
      input_data = f_in.read()

  # convert to ascii
  if type(input_data).__name__ != 'str' :
    input_text =  input_data.decode()
  else:
    input_text = input_data

  assert input_text is not None, "Error of reading input"

  decoded_bytes = base64.b64decode(input_text)
 
  if output == "file":
    with open(output_file, 'wb') as f_out: 
      f_out.write(decoded_bytes)
    return output_file
  elif output == "byte":
    return  decoded_bytes

  pass



####################################################################################################

def reverse_str(text) :
  """
    return reversed the text str, can be used for encryption
    
  """
  reversed_txt = text[::-1]
  return reversed_txt


####################################################################################################

def encrypt(input_file=None,output_file=None):
  """
    encrypt file (has to be coupled with decrypt function)
    no password required.
    
  """
  zip_data = do_zip(input_file=input_file,output="byte")
  b64_data = base64_encode(input_data=zip_data,output="text")
  r_b64_data = reverse_str(b64_data)
  with open(output_file, 'wt') as f_out: 
    f_out.write(r_b64_data)
    f_out.flush()
  pass

####################################################################################################
  
def decrypt(input_file=None,output_file=None):
  """
    decrypt file (has to be coupled with encrypt function)
    no password required.
    
  """
  
  with open(input_file, 'rt') as f_in: 
    r_b64_data = f_in.read()
  b64_data = reverse_str(r_b64_data)
  b64_decoded_data = base64_decode(input_data=b64_data, output="byte") 
  unzip_data =  do_unzip(input_data=b64_decoded_data, output_file=output_file, output="file")
  pass

####################################################################################################


def replace_text_files(replace_keys, files, casesesitive=True):
  """
    usage  : replace text in list of files/single file
    params : 
      oldkeys = old keys to be replaced
      newkeys = new keys to replace old one
      type    = text  
      casesesitive = True (default) | False
    returns : 
      dictionary {} includes files which has been changed     

    sample input :
      result = replace_text_files([('a','a_new'), ('b', 'b_xxx')], ['R:/1.txt','R:/2.txt','R:/3.txt'] )
    
    sample output :
      {
       file1: True,  # file1 changed
       file2: False, # file2 NOT changed
       file3: True   # file3 changed
      }

  """

  result = {}
  if isinstance(files, str) :
    files = [files]

  if isinstance(replace_keys, tuple) :
    replace_keys = [replace_keys]

  for f in files :
    f_text = read_file(f,return_type='str')
    result[f] = False
    for oldkey, newkey in replace_keys:
      if casesesitive == True:
        f_text, count = re.subn(oldkey, newkey, f_text)
      else:
        f_text, count = re.subn(oldkey, newkey, f_text,flags = re.IGNORECASE)
      
      if count > 0 : 
        result[f] = True
        write_file(file_name=f, text = f_text, mode="w")

  return result





if __name__ == "__main__":
  #print (len( get_files (folder="R:/", type='file')))
  """
  for f in gen_files (folder="Y:/important", type='file'):
    print (f)
  

  #print (base64_encode(input_file="R:/panda.jpg",output_file="R:/panda.jpg.b64", output="text"))
  #print (base64_decode(input_file="R:/panda.jpg.b64",output_file="R:/panda1.jpg", output="text"))
  encrypt(input_file="R:/panda.jpg",output_file="R:/panda.jpg.encrypted")
  
  

  decrypt(input_file="R:/panda.jpg.encrypted",output_file="R:/panda_restored.jpg")
  #rm_folder ("R:/", empty=True)
  #print (replace_text_files.__doc__)
  #result = replace_text_files([('http_proxy.*',''), ('b.*', 'BB')], ['R:/1.txt','R:/2.txt','R:/3.txt'],casesesitive=False )

  #print (result)
  
  print(delete_folder.__doc__)
  delete_folder("R:/temp1", empty=False)
  mk_folder("r:/temp1/t1/t2") 
  print (get_file_folder("r:/temp1/t1/t2/t3/file1.txt") )
  print(get_file_name("r:/temp1/t1/t2/t3/file1.txt") )
  print(parse_full_path("r:/temp1/t1/t2/t3/file1.txt") )
  clean_folder("r:/temp1/t1/t2") 
  """
  
  
  """
  split_file(input_file='R:/Temp2/disk.bin',output_folder='R:/Temp2', volume_size=1234)
  join_file(input_file='R:/Temp2/disk.bin', output_folder ='R:/Temp1')
  
  print (get_md5('R:/Temp2/disk.bin'))
  print (get_md5('R:/Temp1/disk.bin'))
  
  """
  """
  clean_folder('R:/TempVM/stage')
  #rm_folder(folder_name='R:/TempVM/stage')
  """
  """
  rm_file(file_name = 'r:/temp1/1.txt') 
  rm_file(file_name = ['r:/temp1/1.txt', 'r:/temp1/3.txt'] ) 
  rm_file(folder = 'R:/temp1', regex_pattern='\d+\.txt' ) 
  """
  
  """
  print (compare_files ( 'R:/TempVM/download/866c6e0a2795e52fc8c99211736a4f70.part_00001', 'R:/TempVM/upload/866c6e0a2795e52fc8c99211736a4f70.part_00000'))
  
  for f in gen_files(folder="R:/Temp2", recursive=True, ext="", type='all'):
    print (f)
  
  for f in get_files(folder="R:/Temp2", recursive=True, ext="", type='F'):
    print (f)
  """
  
  """
  lines = read_file('R:/TempVM/dump.offerings.detail.json', 'list')
  print (len(lines))
  
  exit(1)
  with open('R:/TempVM/dump.offerings.detail.json') as f:
    for line in read_file_lines(f):
       print (line)

  """
   
  
  pass
  


import os,sys
import ctypes

import py3toolbox.text_tools as text_tools

if os.name=='nt' and sys.getwindowsversion().major == 10:
  kernel32 = ctypes.windll.kernel32
  kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

PATTERN = {
  "BRIGHT"      : "1",
  "UNDERLINE"   : "4",
  "DIM"         : "2",

  "WHITE"       : "97",  
  "BLACK"       : "30",
  "RED"         : "31",
  "GREEN"       : "32",
  "BLUE"        : "34",
  "YELLOW"      : "33",
  "MAGENTA"     : "35",
  "CYAN"        : "36",
  
  "LGRAY"       : "37",
  "DGRAY"       : "90",
  "LRED"        : "91",
  "LGREEN"      : "92",
  "LYELLOW"     : "93",
  "LBLUE"       : "94",
  "LMAGENTA"    : "95",
  "LCYAN"       : "96",

  "WHITE_BG"    : "107",  
  "BLACK_BG"    : "40",
  "RED_BG"      : "41",
  "GREEN_BG"    : "42",
  "YELLOW_BG"   : "43",
  "BLUE_BG"     : "44",
  "MAGENTA_BG"  : "45",
  "CYAN_BG"     : "46",

  "LGRAY_BG"    : "47",
  "DGRAY_BG"    : "100",
  "LRED_BG"     : "101",
  "LGREEN_BG"   : "102",
  "LYELLOW_BG"  : "103",
  "LBLUE_BG"    : "104",
  "LMAGENTA_BG" : "105",
  "LCYAN_BG"    : "106",
  
  "DEFAULT"     : 0

}



def cls():
  os.system('cls' if os.name=='nt' else 'clear')


def render256(text, align=None, width=None):
  # regexp : (\[\%(\w+)\|FG(\d+)\|BG(\d+)\|(L|R|\<|\>)\|(\d+)\:(.*?)\%\])
  # sample : [%B|FG233|BG123|L|22:test_color%] and dfgdfg [%B|FG233|BG123|L|22:another test_color%] dfgdfg
  
  render_text = ""
  
  start_flag = '\033['
  end_flag = '\033[0m'
  
  
  ATTR  = ''
  FG    = start_flag + '<ATTR>38;5;<FG>m'
  BG    = start_flag + '48;5;<BG>m'
  
  render_groups = text_tools.re_findall(r'(\[\%(\w+)\|FG(\d+)\|BG(\d+)\|(L|R|M|C|\<|\>)\|(\d+)\:(.*?)\%\])', text)
  
  for render_section in render_groups:
    src_text, attr,  fg_color, bg_color, align, width, raw_text = (render_section[0],render_section[1],render_section[2],render_section[3],render_section[4],render_section[5],render_section[6])
    width = int(width)
    
    
    ATTR  = ''
    if 'B' in attr : ATTR += PATTERN['BRIGHT'] + ';' 
    if 'U' in attr : attr += PATTERN['UNDERLINE'] + ';' 
    if 'D' in attr : attr += PATTERN['DIM'] + ';' 
    
    if align == 'L' : align = '<'
    if align == 'R' : align = '>'
    if align == 'C' or align == 'M' : 
      raw_text = ' ' * ((width - len(raw_text)) // 2) + raw_text + ' ' * (width -  ((width - len(raw_text)) // 2) - len(raw_text) )
    else:
      format_str = '{0:' + align + str(width) + '}'
      raw_text = format_str.format(raw_text)
      
    render_text = FG.replace('<ATTR>',ATTR).replace('<FG>',str(fg_color)) + BG.replace('<BG>',str(bg_color)) + raw_text + end_flag
    
    
    # if not windows 10
    if os.name=='nt' and sys.getwindowsversion().major != 10 :
      render_text = raw_text
      

    text = text.replace(src_text, render_text)

    
  
  return (text)

def render256_demo():
  start_flag = '\033['
  end_flag = '\033[0m '
  FG = start_flag + '38;5;<FG>m'
  BG = start_flag + '48;5;<BG>m'
  count = 0
  render_text = ''
  for bg in range (0,256):
    for fg in range (0,256) :
      count +=1
      text =  BG.replace('<BG>',str(bg)) +  FG.replace('<FG>',str(fg)) + 'bg' + str(bg).rjust(3, '0') + 'fg' + str(fg).rjust(3, '0') + end_flag
      render_text += text 
      if count % 16 == 0:
        print (render_text)
        render_text = ''
    print ('')

  return ""

  
def render(text, align=None, width=None):
  render_groups = text_tools.re_findall(r'(\[\%([\w+\|?]+)\:([^\:\|\%]+)\%\]?)', text)
  render_text   = text
  for render_section in render_groups:
    src_text, render_patterns,raw_text = (render_section[0],text_tools.re_findall(r'(\w+)',render_section[1]),render_section[2])
    term_pattern = '\033['
    for p in render_patterns:
      if p not in PATTERN.keys() : raise ValueError( 'Pattern: [' + p + '] not defined!')
      term_pattern += PATTERN[p] + ';'

    
    # if not windows 10
    if os.name=='nt' and sys.getwindowsversion().major != 10 :
      term_pattern = raw_text
    
    # if windows 10 or *nix
    else:
      term_pattern = term_pattern[:-1] + 'm' + raw_text + '\033[0m'
    
    if (align is not None) and (width is not None):
      if align == '>' or align == 'R' :
        term_pattern = ' ' * (width - len(raw_text)) + term_pattern
      if align == '<' or align == 'L' :
        term_pattern = term_pattern + ' ' * (width - len(raw_text)) 
        
    render_text = render_text.replace(src_text,term_pattern)
  return (render_text)
  

  
  
def pause(txt=None) :
  press_keys_info = "Confirm and press [%LGREEN_BG|BLACK:[Enter]%] to continue ... or [%LRED_BG|BLACK:[Ctrl-C]%] to exit ...\n\n"
  continue_info   = "Pressed [%LGREEN_BG|BLACK:[Enter]%] to continue.\n\n"
  if txt is None : txt = "\n\n---> " 
  pause_txt = txt + press_keys_info
  input(render(pause_txt))
  print(render(continue_info))


def keep_print(text):
  print (text, sep=' ', end='', flush=True)    
  
def get_progress_bar(current, total, scale=50):
  percent = int(current/total*100+0.5)
  if percent <0   : percent = 0
  if percent >100 : percent = 100
  arrow_pos =  int(percent * scale /100 ) 
  if arrow_pos <=1 : 
    progress_bar = '[%LYELLOW:>%]' * arrow_pos + '[%LRED:.%]' * (scale - arrow_pos)
  else:
    if percent < 100 :
      progress_bar = '[%LGREEN:=%]' * (arrow_pos -1) + '[%LYELLOW:>%]' + '[%LRED:.%]' * (scale - arrow_pos)
    else:
      progress_bar = '[%LGREEN:=%]' * scale
  progress_bar = '    [{0}] {1:<5} {2:<12} \r'.format(progress_bar, str(percent) + '%', str(current) + '/' + str(total))
  return render(progress_bar)

def format_str(fmt, *args) :
  return (fmt.format(*args))   

if __name__ == "__main__":
  #'{1} {0}'.format('one', 'two')  
  print(format_str('{1} {0}', 'One', 'Tow'))
  keep_print (get_progress_bar(10,23))
  exit(1)
  _step   =  render('[%LGREEN|DGRAY_BG:12345%] and another [%LGREEN|GREEN_BG:   abcdefgh    %] -',align='<'  ,width=20)
  print (_step)
  _step   =  render('and another [%LGREEN|GREEN_BG:   abcdefgh    %] =',align='<'  ,width=20)
  print (_step)
  
  for i in range (16,21) :
    m = "\033[38;5;" + str(i) + "m#\033[0m"
    print (m)
  #print (render256('[%B|FG184~189|BG46~51|L|22:test_color%] and dfgdfg [%B|FG233|BG123|L|22:another test_color%] dfgdfg'))
  print (render256('[%B|FG014|BG246|L|22:test_color%] and dfgdfg [%U|FG233|BG123|L|22:another test_color%] dfgdfg'))
  print (render256('[%B|FG087|BG235|L|' + str(12) + ':1222222%]'))
  print (render256('[%B|FG087|BG235|C|20:Workers%]'))
  print (render256_demo())
  pass  


MIME_TYPES = {
  "aac"     :	"audio/aac",
  "abw"     :	"application/x-abiword",
  "arc"     :	"application/x-freearc",
  "avi"     :	"video/x-msvideo",
  "azw"     :	"application/vndamazonebook",
  "bin"     :	"application/octet-stream",
  "bmp"     :	"image/bmp",
  "bz"      :	"application/x-bzip",
  "bz2"     :	"application/x-bzip2",
  "csh"     :	"application/x-csh",
  "css"     :	"text/css",
  "csv"     :	"text/csv",
  "doc"     :	"application/msword",
  "docx"    :	"application/vndopenxmlformats-officedocumentwordprocessingmldocument",
  "eot"     :	"application/vndms-fontobject",
  "epub"    :	"application/epub+zip",
  "gz"      :	"application/gzip",
  "gif"     :	"image/gif",
  "htm"     :	"text/html",
  "html"    :	"text/html",
  "ico"     :	"image/vndmicrosofticon",
  "ics"     :	"text/calendar",
  "jar"     :	"application/java-archive",
  "jpeg"    :	"image/jpeg",
  "js"      :	"text/javascript",
  "json"    :	"application/json",
  "jsonld"  :	"application/ld+json",
  "mid"     :	"audio/midi audio/x-midi",
  "mjs"     :	"text/javascript",
  "mp3"     :	"audio/mpeg",
  "mpeg"    :	"video/mpeg",
  "mpkg"    :	"application/vndappleinstaller+xml",
  "odp"		  :	"application/vndoasisopendocumentpresentation",
  "ods"		  :	"application/vndoasisopendocumentspreadsheet",
  "odt"		  :	"application/vndoasisopendocumenttext",
  "oga"		  :	"audio/ogg",
  "ogv"		  :	"video/ogg",
  "ogx"		  :	"application/ogg",
  "opus"		:	"audio/opus",
  "otf"		  :	"font/otf",
  "png"		  :	"image/png",
  "pdf"		  :	"application/pdf",
  "php"		  :	"application/x-httpd-php",
  "ppt"		  :	"application/vndms-powerpoint",
  "pptx"		:	"application/vndopenxmlformats-officedocumentpresentationmlpresentation",
  "rar"		  :	"application/vndrar",
  "rtf"		  :	"application/rtf",
  "sh"		  :	"application/x-sh",
  "svg"		  :	"image/svg+xml",
  "swf"		  :	"application/x-shockwave-flash",
  "tar"		  :	"application/x-tar",
  "tif"		  :	"image/tiff",
  "ts"		  :	"video/mp2t",
  "ttf"		  :	"font/ttf",
  "txt"		  :	"text/plain",
  "vsd"		  :	"application/vndvisio",
  "wav"		  :	"audio/wav",
  "weba"		:	"audio/webm",
  "webm"		:	"video/webm",
  "webp"		:	"image/webp",
  "woff"		:	"font/woff",
  "woff2"	  :	"font/woff2",
  "xhtml"	  :	"application/xhtml+xml",
  "xls"		  :	"application/vndms-excel",
  "xlsx"		:	"application/vndopenxmlformats-officedocumentspreadsheetmlsheet",
  "xml"		  :	"text/xml",
  "xul"		  :	"application/vndmozillaxul+xml",
  "zip"		  :	"application/zip",
  "3gp"		  :	"audio/3gpp",
  "3g2"     :	"audio/3gpp2",
  "7z"      :	"application/x-7z-compressed",
  "folder"  : "application/vnd.google-apps.folder",
  "dir"     : "application/vnd.google-apps.folder",
  "unknown" : "application/unknown"
}


import os
import sys
import re
import smtplib
import email  
import ssl

from email import encoders
from email.mime.multipart import MIMEMultipart 
from email.mime.text import MIMEText 
from email.mime.base import MIMEBase 
from email.mime.image import MIMEImage

def send_mail(config=None):
  """
  sample usage : 
    config = {
        "Subject"     : "This is Mail subject",
        "From"        : "test_12345@tpg.com.au",
        "To"          : "chgpnt@gmail.com",
        "Attachments"  : ["R:/panda.jpg", "R:/tiger.jpg", "R:/baby.jpg"],
        "Body"        : "This is Mail subject",
        "smtp_host"   : "mail.tpg.com.au",
        "smtp_port"   : 465
    }

  """

  msg = MIMEMultipart()
  msg['Subject'] = config["Subject"] 
  msg['From'] = config["From"]
  msg['To'] =  config["To"]

  msg.attach(MIMEText(config["Body"], 'plain'))

  for filename in config["Attachments"]:
    part = MIMEBase('application', "octet-stream")
    part.set_payload(open(filename, "rb").read())
    encoders.encode_base64(part)
    part.add_header('Content-Disposition', 'attachment; filename=' + os.path.basename(filename))
    msg.attach(part)

  context = ssl.create_default_context()
  with smtplib.SMTP_SSL(config["smtp_host"] , config["smtp_port"] , context=context) as server:
    server.sendmail(config["From"], config["To"], msg.as_string())


  pass


def send_gmail (sender_account, sender_password, smtp,smtp_port,senders,receivers,subject,body,attachments):
  msg               = MIMEMultipart()
  msg["From"]       = senders
  msg["To"]         = receivers
  msg["Subject"]    = subject
  
  msg.attach(MIMEText(body, 'plain'))

  for filename in attachments:
    with open(filename, 'rb') as fp:
      part = MIMEBase("application", "octet-stream")
      part.set_payload(fp.read())
    encoders.encode_base64(part)
    attachment_name = os.path.basename(filename)
    part.add_header(
        "Content-Disposition",
        f"attachment; filename={attachment_name}",
    )
    msg.attach(part)

  text = msg.as_string()


  context = ssl.create_default_context()
  with smtplib.SMTP_SSL(smtp, smtp_port, context=context) as server:
    server.login(sender_account, sender_password)
    server.sendmail(senders, receivers, text)



if __name__ == "__main__":
  config = {
      "Subject"     : "test_icc_group_email",
      "From"        : "test_icc_group_email@gmail.com",
      "To"          : "fan.yang36@tafensw.edu.au",
      "Attachments" : [],
      "Body"        : "This is a test of ICC group email",
      "smtp_host"   : "mail.tpg.com.au",
      "smtp_port"   : 465
  }
  send_mail(config)
  """
  send_gmail('xxxxxx@gmail.com',
            'xxxxxx',
            'smtp.gmail.com', 
            465, 
            'xxxxx@gmail.com',
            'xxxxx@test.com',
            'System Monitor ' ,
            'blablabla',
            ["R:/car.jpg", "R:/car.rar", "R:/1.cat"]
            )
  """


from __future__ import print_function


import pickle, io
import os.path
import copy
from py3toolbox import fs_tools   as fs_tools
from py3toolbox import helper     as helper




# ====================================================================
#
# This google drive api wrapper is based on Goodgle Drive API v3
#
# ====================================================================

from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request



SCOPES          = ['https://www.googleapis.com/auth/drive']
TOKEN_PICKLE    = 'gdrive.<ID>.token.pkl'
CHUNK_SIZE      = 100*1024*1024


class Gdrive():
  def __init__(self, cred_file=None, cred_id=1, page_size=100):
    self.cred_file          = cred_file
    self.cred_id            = str(cred_id)
    self.token_pickle       = TOKEN_PICKLE.replace('<ID>', self.cred_id)
    self.page_size          = page_size
    self.meta_cache         = None
    self.gdrive_service_v3  = None
    self.auth_cred()
    self.init_gdrive()
    
    pass
    
    
    
  # ====================================================
  #
  # authenticate login
  #
  # ====================================================    
  def auth_cred(self):
    """
      authenticate user credential
    
    
    """
    
    creds = None
    if os.path.exists(self.token_pickle):
      with open(self.token_pickle, 'rb') as token:
        creds = pickle.load(token)
    
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
      if creds and creds.expired and creds.refresh_token:
        creds.refresh(Request())
      else:
        flow = InstalledAppFlow.from_client_secrets_file(self.cred_file, SCOPES)
        creds = flow.run_local_server(port=0)
      
      # Save the credentials for the next run
      with open(self.token_pickle, 'wb') as token:
        pickle.dump(creds, token)

    assert creds is not None, "Gdrive credential error."
    
    self.gdrive_service_v3 = build('drive', 'v3', credentials=creds)
    
    pass
    


  # ====================================================
  #
  # init gdrive
  #
  # ====================================================    
    
  def init_gdrive(self):
    self.root_id  = self.gdrive_service_v3.files().get(fileId="root", fields=("id")).execute()['id']  
    pass
    

  # ====================================================
  #
  # Parse full names of the given name (path)
  #
  # ====================================================
  
  def _parse_full_name (self, name):
  
    if name is None:
      return []
      
    name  = name.replace('\\','/').replace('//','/')

    while name[:1] == '/' : 
      name = name[1:]
      
    names = name.split('/')
    return names    


  # ====================================================
  #
  # Apply operation task on cache hierachy
  #
  # ====================================================

  def _get_hierchy(self, tree_data, id, operation="list" ) :
    id_collections = []
    tree_data_result = {}

    # find out the impacted id list
    while True:
      v = len(id_collections)
      for k in tree_data.keys() :
        r = tree_data[k]
        parent_id = r['parents'][0]
        if k == id and k not in id_collections:
          id_collections.append(k)
          if (parent_id != self.root_id) and parent_id not in id_collections :
            id_collections.append(parent_id)  
          
        if parent_id in id_collections and k not in id_collections:
          id_collections.append(k)
          
      id_collections = list(set(id_collections))
      
      if v == len(id_collections) :
        break;
    
    
    # execute the operation task
    if operation == "list" :
      for k in tree_data.keys() :
        r = tree_data[k] 
        if k in id_collections:
          tree_data_result[k] = r
     
     
    elif operation == "remove":
      tree_data_result = copy.deepcopy(tree_data)
      for k in tree_data.keys() :
        if k in id_collections:
          tree_data_result.pop(k, None)
  
    return id_collections,tree_data_result
      

  # ====================================================
  #
  #  download files from gdrive and verify MD5
  #
  # ====================================================
  
  def download(self, id, localname, verify=False):
    download_request = self.gdrive_service_v3.files().get_media(fileId=id)
    file_handler = io.FileIO(localname, mode='wb')
    downloader = MediaIoBaseDownload(fd=file_handler, request=download_request, chunksize=CHUNK_SIZE)
    done = False
    try :
      while done is False:
        status, done = downloader.next_chunk()
    except Exception as err:
      return False, "download:" + str(err)

    if verify :
      src_md5 = self.get_meta(id=id)['md5Checksum']
      dst_md5 = fs_tools.get_md5(localname)
      if src_md5==dst_md5:
        return True, "success"
      else:
        return False, "Downloaded, but verification failed. src_md5=" + src_md5 + ", dst_md5=" + dst_md5
    else:
      return True, "success"
    


  def upload(self, id, localname, verify=False):
    
    src_md5 = None
    dst_md5 = None
    
    assert fs_tools.exists(localname), "local file doesn't exist"
    
    
    if verify:
      src_md5 = self.get_meta(id=id)['md5Checksum']
      dst_md5 = fs_tools.get_md5(localname)
      if src_md5 == dst_md5:
        return True, "upload: bypassed."
    
    try :
      body_meta = {
        "mimetype" : helper.get_mimetype(localname=localname)
      }
      file_upload = MediaFileUpload(filename=localname, mimetype=helper.get_mimetype(localname=localname), chunksize=CHUNK_SIZE )
      file = self.gdrive_service_v3.files().update(fileId=id, body=body_meta, media_body=file_upload).execute()
    except Exception as err:
       return Fasle, "upload:" + str(err)

    if verify:
      src_md5 = self.get_meta(id=id)['md5Checksum']
      if src_md5 == dst_md5:
        return True, "upload: verified"
    else:
      return True, "upload: success"
    
    
    
    
    
  # ====================================================
  #
  #  Create file
  #
  # ====================================================    
    
  def create(self, name, localname):
    
    names=self._parse_full_name(name)
    
    if len(names) == 1: 
      parent = self.root_id
    else:
      pass
    
  
    # if file not exist, then create
    # otherwise : TODO
    if len(self.get_id(name=name)) == 0: 
      type = helper.get_mimetype(type=type)
      
      file_metadata = {
        'name': name,
        'mimeType': type,
        'parents' : [ parent ]
      }
      
      file = self.gdrive_service_v3.files().create(body=file_metadata, fields='id').execute()  
    else:
      return None
      
    
    
    if localname is not None and type != helper.get_mimetype(type="folder") :
      self.upload(localname=localname )
    
    return file.get('id')
  


  # ====================================================
  #
  #  Gdrive file list generator
  #
  # ====================================================    

  def gen_list(self, name=None, is_folder=None, parent=None, cust_properties=None, fields=None, order_by='folder,name,createdTime'):
    q = []
    params = {'fields' : fields, 'pageSize' : self.page_size, 'pageToken': None, 'orderBy': order_by}
    

    if fields is not None:
      params['fields'] =  ','.join([ 'files/'+ f.replace('files/','')  for f in fields.split(',')])
    else:
      params.pop('fields', None)
    
    if name is not None:
      q.append("name = '%s'" % name.replace("'", "\\'"))
    if is_folder is not None:
      q.append("mimeType %s '%s'" % ('=' if is_folder else '!=', helper.get_mimetype(type='folder')))
    if parent is not None:
      q.append("'%s' in parents" % parent.replace("'", "\\'"))
    if cust_properties is not None:
      for k,v in cust_properties.items():
        q.append("properties has { key='%s' and value='%s' }" % (k , v.replace("'", "\\'") ) )
      
    
    # add query expression
    if q: params['q'] = ' and '.join(q)

    while True:
      response = self.gdrive_service_v3.files().list(**params).execute()
      for f in response['files']:
        yield f
      try:
        params['pageToken'] = response['nextPageToken']
      except KeyError:
        return
    
    
    
  # ====================================================
  #
  #  Gdrive file list 
  #
  # ====================================================    
  
  def get_list(self,**kwargs):
    result = []
    response = self.gen_list(**kwargs)
    for f in response:
      result.append(f)
    
    return result
  

  # ===============================================================  
  #  Get Full name (path) of the GDrive file
  #  For example: /folder1/sub-folder2/sub_folder3/the_file
  #
  #
  # ===============================================================
  def get_id (self, name=None, id=None, force_update_cache=False):
    assert name is not None or id is not None , "provide at least one of name or id"
    
    if force_update_cache == True or self.meta_cache is None:
      self.get_meta_cache(name=name, id=id)
      
    ids = []
    
    if name is not None:
      for k in self.meta_cache.keys():
        r = self.meta_cache[k]
        if r['full_name'][-len(name):] == name :
          ids.append(k)
          
    else:
      for k in self.meta_cache.keys():
        r = self.meta_cache[k]
        if k == id :
          ids.append(k)
      
    return ids
      
  

  # ===============================================================  
  #  Get Full name (path) of the GDrive file
  #  For example: /folder1/sub-folder2/sub_folder3/the_file
  #
  #
  # ===============================================================
  def get_full_name (self, name=None, id=None, force_update_cache=False):
    assert name is not None or id is not None , "provide at least one of name or id"
    
    if force_update_cache == True or self.meta_cache is None:
      self.get_meta_cache(name=name, id=id)
      
    full_name = []
    
    if name is not None:
      for k in self.meta_cache.keys():
        r = self.meta_cache[k]
        if r['full_name'][-len(name):] == name :
          full_name.append(r['full_name'])
          
    else:
      for k in self.meta_cache.keys():
        r = self.meta_cache[k]
        if k == id :
          full_name.append(r['full_name'])
      
    return full_name
  
  

  def delete(self, id, trash=False):
    if self.get_meta(id=id) is not None:
      try :
        if trash :
          self.gdrive_service_v3.files().trash(fileId=id).execute()
        else:
          self.gdrive_service_v3.files().delete(fileId=id).execute()
      except Exception as err:
        return False, "delete:" + str(err)
    else:
      return True, "success"
  

  
  def set_meta(self,id,body_meta):
    try :
      results = self.gdrive_service_v3.files().update(fileId=id, body=body_meta).execute()
    except Exception as err:
      return False, "set_meta:" + str(err)
    return True, "success"
    

  def get_meta(self,id,fields="*"):
    results = None
    try :
      results = self.gdrive_service_v3.files().get(fileId=id, fields=fields).execute()
    except Exception as err:
      pass
    return results



  # ============================================================
  #
  # Read meta cache of the file and parant chain.
  # 
  #
  # ============================================================
  
    
  def get_meta_cache(self, name=None, id=None, tree=True):
    assert name is not None or id is not None , "provide at least one of name or id"
    
    layer_meta = {}
    chain_meta = {}
    names = self._parse_full_name(name)
    
    if name is not None:
      n = names[-1]
      found_list = self.get_list(name=n)
      for f in found_list:
        id = f['id']
        layer_meta[id] = self.get_meta(id=id)
    else:
      layer_meta[id] = self.get_meta(id=id)
      
    while True:
      chain_meta = copy.deepcopy(layer_meta)
      
      names = names[:-1]
      
      if len(names) >0 :
        n = names[-1]
      else:
        n = None
      
      for k in layer_meta.keys():
        r = layer_meta[k]
        parent_id = r['parents'][0]
        
        if parent_id == self.root_id :
          continue
        
        if parent_id in chain_meta :
          continue
       
        t = self.get_meta(id=parent_id)

        # if parent name match the n
        if t['name'] == n or n is None and tree==True:
          chain_meta[parent_id] = t
          
        # if parent name not match the n, then remove the node meta
        if t['name'] != n and n is not None :
          _, chain_meta = self._get_hierchy (tree_data = chain_meta, id=k, operation="remove")
      
      if chain_meta != layer_meta:
        layer_meta = copy.deepcopy(chain_meta)
      else:
        break
    
    # if no entire tree to be returned
    
    if tree == False:
      self.meta_cache = chain_meta
      return self.meta_cache
    
    
    # cleanup layer_meta
    clean_chain_meta = {}
    while True:
      v = len(clean_chain_meta.keys())
      for k in layer_meta.keys():
        
        r = layer_meta[k]
        parent_id = r['parents'][0]
        
        if parent_id == self.root_id :
          clean_chain_meta[k] = r
          clean_chain_meta[k]["full_name"] = "/" + clean_chain_meta[k]["name"]
      
        
        if parent_id in clean_chain_meta :
          clean_chain_meta[k] = r
          clean_chain_meta[k]["full_name"] = clean_chain_meta[parent_id]["full_name"] + "/" + clean_chain_meta[k]["name"] 
        
      if v == len(clean_chain_meta.keys()):
        break;
      
    self.meta_cache = clean_chain_meta
    return self.meta_cache



if __name__ == "__main__":
  my_cloud = Gdrive(cred_file="R:/credentials.json" )
  #my_cloud.get_folder_list()
  
  #for f in my_cloud.search_folder(name= "root"):   print (f)
  #print (my_cloud.get_id(name="/test2/test2/1_junco.jpg"))
  
  #print (my_cloud.get_meta_cache(id="10KGzUTKURK82_OU-1qFuxYKdgUKHD4LC"))
  
  #print (my_cloud.get_meta_cache(name="/test2/1_junco.jpg"))
  #print (my_cloud.get_full_name(name="1_junco.jpg"))
  #print (my_cloud.get_full_name(id="10KGzUTKURK82_OU-1qFuxYKdgUKHD4LC"))
  #print (my_cloud.get_id(name="1_junco.jpg"))

  #print (my_cloud.is_folder(name="1_junco.jpg"))
  #print (my_cloud.create(name="LP6.txt"))
  
  #print(helper.get_mimetype(localname="R:/IMG1.bin"))
  #print (my_cloud.get_meta_cache(name="LP6.txt", tree=False))
  #print (my_cloud.upload(id="1KGmskL2N8f7snWP_EvLTKyzpz4uJ1RTp", localname="R:/IMG1.bin", verify=True))
  
  body_meta = { 

                "description" : " { \"archive\": \" abc\" } ",
                "properties" : { 
                  "my_prop" : "test_p1",
                  "src_name_key" : "abcdefg"
                } ,
                
                "contentHints" : {
                  "indexableText" : " ___ a long description here ___"
                
                
                }
              }
              
  #print(my_cloud.set_meta(id="1KGmskL2N8f7snWP_EvLTKyzpz4uJ1RTp", body_meta=body_meta))
  
  cust_properties =  {'my_prop': 'test_p1' }
  
  #for f in my_cloud.get_list(cust_properties = cust_properties) :
  #  print (f)
    
    
  print (my_cloud.get_list(cust_properties = cust_properties, fields="kind,id,name,modifiedTime,properties/src_name_key,md5Checksum") )
  #print(my_cloud.download(id="1nAD4tlsAIAM2Cadg3LqY1DlB5CoZrZrJ", localname="R:/t1.dat", verify=True))
  #print(my_cloud.delete(id="1nAD4tlsAIAM2Cadg3LqY1DlB5CoZrZrJ"))
  
  
  
  
  
  """
  f = my_cloud.get_list(name="test2")
  print (f)
  print (my_cloud.get_parent_id(f[0]['id']))
  #my_cloud.delete_file(id='1BTVsE1jU0X9cwPP88xElW90FUHZTvBPm')

    
  #14_i-9jqqzPC6sECZwVH-IxTKrD93ueC7
  #0ALjvPRMe8hRyUk9PVA
  print (my_cloud.get_full_name(name='1_rangandas-gardenglimpses4.jpg'))
  print (my_cloud.create(name='my_file2.txt'))
  #print (my_cloud.get_name(id='0ALjvPRMe8hRyUk9PVA'))  
  
  #my_cloud.search_files_folers()
  #my_cloud.create_folder(folder_name="test2")
  #print (my_cloud.get_id(name='test1'))
  #print (my_cloud.root_id)
  """



import os,sys


def gen_css(config):
  pass

def gen_table_head(config):
  
  return ''

def gen_table_row(config):
  
  return ''
  
def gen_table_cell(config):
  
  return ''  

def gen_table_tail(config):
  
  return ''    

def gen_html_head(config):
  
  return ''      
  
def gen_html_tail(config):
  
  return ''        

def gen_button(config):
  
  return ''        


import xml.etree.ElementTree as et

def is_xml(xml_text) :
  try :
    x = et.fromstring(xml_text)
    return True
  except Exception as e:
    return False  


def pretty_xml(xml_file) :
  xml_doc = et.parse(xml_file)
  return (et.tostring(xml_doc, pretty_print=True).decode("utf-8")) 
  
  
def parse_xml(xml_text) :
  try :
    xml_obj = et.fromstring(xml_text)
    return xml_obj
  except Exception as e:
    return None  


def get_xml_elements(xml_obj,xpath,value=None):
  elements_found = []
  xml_elements = xml_obj.findall(xpath)
  for e in xml_elements:
    if value is not None and value == e.text:
      elements_found.append(e)
    if value is None: 
      elements_found.append(e)
  return elements_found    

  
def get_xml_value(xml_obj,xpath):
  xml_element = None
  xml_element = xml_obj.find(xpath)
  if xml_element is not None : 
    return xml_element.text
  else :
    return ""

def get_xml_values(xml_obj,xpath):
  xml_values = []
  xml_elements = xml_obj.findall(xpath)
  for e in xml_elements:
    xml_values.append(e.text)
  return xml_values

  
def get_xml_attributes(xml_obj, xpath, attrib):
  xml_attributes = []
  xml_elements = xml_obj.findall(xpath)
  for e in xml_elements:
    xml_attributes.append(e.get(attrib))
  return xml_attributes

def contain_xml_value(xml_obj, xpath, value):
  values = get_xml_values(xml_obj, xpath)
  if value in values: return True
  return False

def contain_xml_attribute(xml_obj, xpath, attrib, value):
  attrib_values = get_xml_attributes(xml_obj, xpath, attrib)
  if value in attrib_values: return True
  return False

  
if __name__ == "__main__":
  import fs_tools
  xml = parse_xml(fs_tools.read_file('C:/TEMP/test.xml'))
  print (xml)
  #print (get_xml_value(xml, "./LearningProductOrganisationOffering/LearningProductOrganisationOffering-LearningProductOrganisationOfferingRelationshipList/LearningProductOrganisationOffering-LearningProductOrganisationOfferingRelationship/LearningProductOrganisationOfferingTo/LearningProduct/ProductType/Code"))
  
  print ( get_xml_values(xml, "./country[rank='4']/gdppc"))

  
  pass    


import os
import re
 
def re_search (re_pattern , text) :
  m = re.search(re_pattern, text)
  if m :  return m.groups
  return None

  
def re_findall (re_pattern , text) :
  m = re.findall(re_pattern, text)
  return m

def re_sub(re_pattern , text):
  return re.sub(re_pattern, text)


def map_none(input=None):
  if input is not None:
    return input
  else:
    return ""

  
  
if __name__ == "__main__":
  str1 = 'product/900/162-11292V02G01'
  a = re_search('([^\/]*)$',str1)
  print (a(0)[0])
  #b = re_findall ('yaw=(\-*\d+\.\d*),pitch=(\-*\d+\.\d*),roll=(\-*\d+\.\d*)', 'MPU_1,yaw=-29.47,pitch=-7.54,roll=0.91')
  b = re_findall ('ENC\((.*)\)', 'truststore.password=EaNC(/ndgPwyPzRc56sv7r0BkSbGip76SN7GVhQqeyOzmQ+Y=)')
  print (b)
  
  pass  


import json
#import .fs_tools as fs_tools

def is_json(json_text):
  try:
    json_object = json.loads(json_text)
  except ValueError:
    return False
  return True    
  
def parse_json_text(body=None, return_fmt="dict"):
    assert return_fmt in ["dict", "text"], f"return_fmt [{return_fmt}] has to be either dict or text"
    
    if body is None: return body
    
    body_data = None
    body_text = None
    
    if type(body).__name__ == "str":
      assert is_json(body) == True, f"body is not json text"
      body_data = json.loads(body)
      body_text = pretty_json(body)
    elif type(body).__name__ == "dict":
      body_data = body
      body_text = pretty_json(body)      
    else:
      assert 1==0, f"invalid body type : {type(body).__name__}"
    
    if return_fmt == "dict" : return body_data
    if return_fmt == "text" : return body_text
      


def load_json(json_file):
  with open(json_file , encoding='utf-8') as json_fh:
    config = json.load(json_fh)
  return config    


def pretty_json(json_input) :
  if type(json_input).__name__ == "str" :
    return (json.dumps(json.loads(json_input), sort_keys=True, indent=2))
  
  if type(json_input).__name__ == "dict" :
    return (json.dumps(json_input, sort_keys=True, indent=2))
  
if __name__ == "__main__": 
  a  = '{"a":6}'
  a  = {"a":6}
  print (parse_json_text(body=a, return_fmt="text"))
  
  
  print 
  pass


#ffmpeg -i INFILE.mp4 -vcodec copy -acodec copy -ss 00:00:00 -t 00:00:10.000 OUTFILE.mp4
import os
import sys
import subprocess
import re
import time
import os

def get_video_duration(src_file):
  p = subprocess.Popen([
    "C:/Tools/ffmpeg/bin/ffprobe.exe", 
    "-v", "error",
    "-show_entries",
    "format=duration",
    "-of",
    "default=noprint_wrappers=1:nokey=1",
    src_file
  ], stdout=subprocess.PIPE)
  duration = int(float(p.communicate()[0].decode("utf-8")))
  return duration


def split_by_time(time_interval, duration):
  file_index = 0
  start_time = 0
  end_time   = 0
  split_list = []
  
  while (True):
    file_index +=1
    if (start_time + time_interval > duration) :
      end_time = duration
    else:
      end_time = start_time + time_interval

    split_list.append({"index":file_index, "start":start_time, "end" : end_time})
    if end_time >= duration : break
    start_time = end_time
    
  return split_list
  
def split_video(source_name, time_interval):
  #time.strftime('%H:%M:%S', time.gmtime(12345))
  #ffmpeg -i source.m4v -ss 1144.94 -t 581.25 -c copy part3.m4v
  filename, file_extension = os.path.splitext(source_name)
  split_list = split_by_time(time_interval,get_video_duration(source_name))
  split_part_list = []
  for part in split_list:
    part["output"] = filename + "_{:04}".format(part["index"]) + file_extension
    if os.path.isfile(part["output"]):  os.remove(part["output"])
    split_part_list.append(part)
  
  for part in split_part_list:
    p = subprocess.Popen([
      "C:/Tools/ffmpeg/bin/ffmpeg.exe",
      "-loglevel",
      "panic",
      "-i",
      source_name,
      "-ss",
      str(part["start"]),
      "-to",
      str(part["end"]),
      "-c",
      "copy",
      part["output"]
    ], stdout=subprocess.PIPE)    
    
  return split_part_list  

if __name__ == "__main__":
  print (split_video("R:/x.mp4",360))
   
  


  
 



__author__  = "Great Tomorrow(gr82morozr@gmail.com)"
__license__ = "MIT"
__version__ = "0.1.39",

from .fs_tools        import  *
from .img_tools       import  *
from .net_tools       import  *
from .print_tools     import  *
from .time_tools      import  *
from .json_tools      import  *
from .xml_tools       import  *
from .text_tools      import  *
from .os_tools        import  *
from .html_tools      import  *
from .mail_tools      import  *
from .dict_tools      import  *
from .media_tools     import  *
from .daemon_tools    import  *
#from .mouse_tools     import  *
from .helper          import  *




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name          = 'py3toolbox',
    version       = '0.1.39',
    author        = 'Great Tomorrow',
    author_email  = 'gr82morozr@gmail.com',
    description   = 'A Python3 tools and utilities collection',
    licence       = 'MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    
    url           = 'https://github.com/gr82morozr/py3toolbox.git',  
    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],

    keywords    = 'toolbox',
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires = ['exifread', 'cryptography', 'pyautogui','numpy']

)


import os
import sys
import time
import random

from py3toolbox import constants  as con


def get_mimetype(type=None, localname=None) :
  assert type is not None or localname is not None,  "type and localname, supply at least one parameter"
  
  if type is None: 
    type = localname.split(".")[-1]
  
  
  if type not in con.MIME_TYPES :
    type = "unknown"
  
  
  return con.MIME_TYPES[type]
    


def get_uid(case='L', digits=8):
  if case=='L' :
    alphabet = 'abcdefghijklmnopqrstuvwxyz0123456789'
  elif case == 'U':
    alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
  else:
    alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
  return ''.join(random.choices(alphabet, k=8))
    
  
if __name__ == "__main__":
  print (get_mimetype(type="folder"))
  print (get_mimetype(localname="r:/cow.png"))
  print (get_mimetype())
  


"""
  ===== Reference ====
  
  URL: https://docs.python.org/3/library/time.html
  
  %a -> Locale’s abbreviated weekday name.
  %A -> Locale’s full weekday name.

  %b -> Locale’s abbreviated month name.
  %B -> Locale’s full month name.

  %c -> Locale’s appropriate date and time representation.
  %d -> Day of the month as a decimal number [01,31].
  %H -> Hour (24-hour clock) as a decimal number [00,23].
  %I -> Hour (12-hour clock) as a decimal number [01,12].
  %j -> Day of the year as a decimal number [001,366].

  %m -> Month as a decimal number [01,12].
  %M -> Minute as a decimal number [00,59].

  %p -> Locale’s equivalent of either AM or PM.
  %S -> Second as a decimal number [00,61].
  %U -> Week number of the year (Sunday as the first day of the week) as a decimal number [00,53]. All days in a new year preceding the first Sunday are considered to be in week 0.

  %w -> Weekday as a decimal number [0(Sunday),6].
  %W -> Week number of the year (Monday as the first day of the week) as a decimal number [00,53]. All days in a new year preceding the first Monday are considered to be in week 0.

  %x -> Locale’s appropriate date representation.
  %X -> Locale’s appropriate time representation.

  %y -> Year without century as a decimal number [00,99].
  %Y -> Year with century as a decimal number.

  %z -> Time zone offset indicating a positive or negative time difference from UTC/GMT of the form +HHMM or -HHMM, where H represents decimal hour digits and M represents decimal minute digits [-23:59, +23:59].
  %Z -> Time zone name (no characters if no time zone exists).

  %% -> A literal '%' character.

"""

import time
import datetime
import math


LOCAL_TIMEZONE            = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
LOCAL_TIMEZONE_OFFSET     = - time.timezone
DEFAULT_TIMESTAMP_FORMAT  = "%Y-%m-%d %H:%M:%S"
ISO_TIMESTAMP_FORMAT      = "%Y-%m-%dT%H:%M:%S"




def get_timestamp(fmt=DEFAULT_TIMESTAMP_FORMAT, epoch_time = None, iso_format=False , timezone=LOCAL_TIMEZONE) :
  assert timezone==LOCAL_TIMEZONE or  timezone=="UTC", "Only support UTC or local timezone"
  if epoch_time is not None : 
    digits = int(math.log10(epoch_time))+1
    if digits > 10 :  epoch_time = int(str(epoch_time)[:10])
  else:
    epoch_time = time.time()

  if timezone == "UTC" :
    time_struct = 	time.gmtime(epoch_time)
  else:
    time_struct = 	time.localtime(epoch_time)
  
  if iso_format :
    fmt = ISO_TIMESTAMP_FORMAT
    
  return  (time.strftime(fmt, time_struct))


def parse_timestamp(ts_str, fmt=DEFAULT_TIMESTAMP_FORMAT, timezone=LOCAL_TIMEZONE):
  # timezone is the timezone of input ts_str, NOT the output
  time_struct   = time.strptime(ts_str, fmt)
  epoc_time     = time.mktime(time_struct)
  
  if timezone == "UTC" :
    epoc_time = epoc_time + LOCAL_TIMEZONE_OFFSET
    
  return int(epoc_time)



def timer_start():
  start = time.time()
  return (start)  
  
def timer_check(start):
  return(time.time() - start)   


def readable_duration(seconds):
  day = seconds // (24 * 3600)
  seconds %= (24 * 3600)
  
  hour = seconds // 3600
  seconds %= 3600
  
  minutes = seconds // 60
  seconds %= 60
  
  return(day, hour, minutes, seconds)

def cal_ts_diff(ts1=None, ts2=None):
  assert ts1 is None or type(ts1).__name__ == "int", 'Timestamp ts1 needs to be in epoch seconds.'
  assert ts2 is None or type(ts2).__name__ == "int", 'Timestamp ts1 needs to be in epoch seconds.'
  
  if ts1 is None:   ts1 = time.time()
  if ts2 is None:   ts2 = time.time()

  
  return ts2 - ts1




if __name__ == "__main__":
  #print (get_timestamp(fmt=DEFAULT_TIMESTAMP_FORMAT, epoch_time=1579046194))
  #print (get_epoch(get_timestamp()))
  #print (time.time())
  
  #print (parse_ts_str('2020-07-02T02:57:24.331Z'))
  #print (parse_ts_str(ts_str='2020-07-02T02:57:24.331Z', fmt="%Y-%m-%dT%H:%M:%S.%fZ"))
  #dt1= parse_ts_str(ts_str='2020-07-02T02:57:24.331Z', fmt="%Y-%m-%dT%H:%M:%S.%fZ")
  
  #print (cal_days_diff(dt1='2020-07-02T02:57:24.331Z',fmt="%Y-%m-%dT%H:%M:%S.%fZ")) 
  
  
  
  #LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone()
  #print (LOCAL_TIMEZONE)
  #LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone(datetime.timedelta(0))).astimezone().tzinfo
  #print (LOCAL_TIMEZONE)
  
  #ts = '2020-07-02T02:57:24.331Z'
  #print (time.time())
  #print (get_timestamp())
  #print (get_timestamp(timezone="UTC"))
  
  #print(parse_timestamp(ts_str=ts, fmt="%Y-%m-%dT%H:%M:%S.%fZ", timezone="UTC"))

  #print (cal_ts_diff(ts2=parse_timestamp(ts_str=ts, fmt="%Y-%m-%dT%H:%M:%S.%fZ", timezone="UTC")))
  print (readable_duration(120202))
  pass
  


import os
import subprocess

def exec_sys_cmd(sys_cmd):
  sp = subprocess.Popen(sys_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  out, err = sp.communicate()
  return (out.decode('utf-8'), err.decode('utf-8'))
  


def boot_remote_host(wol_cmd, local_nic, remote_host, remote_host_mac):
  boot_cmd    = [wol_cmd, '-i', local_nic , remote_host_mac] 

  for i in range (int(me.config['REMOTE_BOOT_MAX_TIME'])):
    time.sleep(1)
    output,error = me.exec_sys_cmd(detect_cmd)
    me.write_log(log_file, '[' + me.config['REMOTE_HOST'] + '] : ' + output + error)
    if me.config['REMOTE_HOST'] in output: 
      me.write_log(log_file, 'Remote Host Started :' + me.config['REMOTE_HOST'])
      return
   
  
if __name__ == "__main__":
  print (exec_sys_cmd("cmd"))
  pass  
  